@@ -19,16 +19,19 @@ Use these classes as values for the `type` argument to
 
 .. versionadded:: 1.3
 """
+import collections
 import operator
 import re
 import warnings
 
 import abc
+from debtcollector import removals
 import netaddr
-import six
+import rfc3986
 
 
-class ConfigType(object):
+class ConfigType(object, metaclass=abc.ABCMeta):
+
     def __init__(self, type_name='unknown type'):
         self.type_name = type_name
 
@@ -39,7 +42,10 @@ class ConfigType(object):
 
         """
         if sample_default is not None:
-            default_str = sample_default
+            if isinstance(sample_default, str):
+                default_str = sample_default
+            else:
+                default_str = self._formatter(sample_default)
         elif default is None:
             default_str = self.NONE_DEFAULT
         else:
@@ -47,9 +53,9 @@ class ConfigType(object):
         return [default_str]
 
     def quote_trailing_and_leading_space(self, str_val):
-        if not isinstance(str_val, six.string_types):
+        if not isinstance(str_val, str):
             warnings.warn('converting \'%s\' to a string' % str_val)
-            str_val = six.text_type(str_val)
+            str_val = str(str_val)
         if str_val.strip() != str_val:
             return '"%s"' % str_val
         return str_val
@@ -65,8 +71,8 @@ class String(ConfigType):
 
     String values do not get transformed and are returned as str objects.
 
-    :param choices: Optional sequence of valid values. Mutually
-                    exclusive with 'regex'.
+    :param choices: Optional sequence of either valid values or tuples of valid
+        values with descriptions. Mutually exclusive with 'regex'.
     :param quotes: If True and string is enclosed with single or double
                    quotes, will strip those quotes. Will signal error if
                    string have quote at the beginning and no quote at
@@ -93,6 +99,10 @@ class String(ConfigType):
     .. versionchanged:: 2.7
        Added *max_length* parameter.
        Added *type_name* parameter.
+
+    .. versionchanged:: 5.2
+       The *choices* parameter will now accept a sequence of tuples, where each
+       tuple is of form (*choice*, *description*)
     """
 
     def __init__(self, choices=None, quotes=False, regex=None,
@@ -106,17 +116,24 @@ class String(ConfigType):
         self.quotes = quotes
         self.max_length = max_length or 0
 
-        self.choices = choices
+        if choices is not None:
+            if not all(isinstance(choice, tuple) for choice in choices):
+                choices = [(choice, None) for choice in choices]
+
+            self.choices = collections.OrderedDict(choices)
+        else:
+            self.choices = None
+
         self.lower_case_choices = None
         if self.choices is not None and self.ignore_case:
-            self.lower_case_choices = [c.lower() for c in choices]
+            self.lower_case_choices = [c.lower() for c in self.choices]
 
         self.regex = regex
         if self.regex is not None:
             re_flags = re.IGNORECASE if self.ignore_case else 0
 
             # Check if regex is a string or an already compiled regex
-            if isinstance(regex, six.string_types):
+            if isinstance(regex, str):
                 self.regex = re.compile(regex, re_flags)
             else:
                 self.regex = re.compile(regex.pattern, re_flags | regex.flags)
@@ -125,7 +142,7 @@ class String(ConfigType):
         value = str(value)
         if self.quotes and value:
             if value[0] in "\"'":
-                if value[-1] != value[0]:
+                if len(value) == 1 or value[-1] != value[0]:
                     raise ValueError('Non-closed quote: %s' % value)
                 value = value[1:-1]
 
@@ -143,7 +160,7 @@ class String(ConfigType):
         # Check for case insensitive
         processed_value, choices = ((value.lower(), self.lower_case_choices)
                                     if self.ignore_case else
-                                    (value, self.choices))
+                                    (value, self.choices.keys()))
         if processed_value in choices:
             return value
 
@@ -154,8 +171,8 @@ class String(ConfigType):
 
     def __repr__(self):
         details = []
-        if self.choices:
-            details.append("choices=%r" % self.choices)
+        if self.choices is not None:
+            details.append("choices={!r}".format(list(self.choices.keys())))
         if self.regex:
             details.append("regex=%r" % self.regex.pattern)
         if details:
@@ -165,11 +182,12 @@ class String(ConfigType):
     def __eq__(self, other):
         return (
             (self.__class__ == other.__class__) and
-            (set(self.choices) == set(other.choices) if
-             self.choices and other.choices else
-             self.choices == other.choices) and
             (self.quotes == other.quotes) and
-            (self.regex == other.regex)
+            (self.regex == other.regex) and
+            (set([x for x in self.choices or []]) ==
+                set([x for x in other.choices or []]) if
+             self.choices and other.choices else
+             self.choices == other.choices)
         )
 
     def _formatter(self, value):
@@ -241,7 +259,100 @@ class Boolean(ConfigType):
         return str(value).lower()
 
 
-class Integer(ConfigType):
+class Number(ConfigType):
+
+    """Number class, base for Integer and Float.
+
+    :param num_type: the type of number used for casting (i.e int, float)
+    :param type_name: Type name to be used in the sample config file.
+    :param min: Optional check that value is greater than or equal to min.
+    :param max: Optional check that value is less than or equal to max.
+    :param choices: Optional sequence of either valid values or tuples of valid
+        values with descriptions.
+
+    .. versionadded:: 3.14
+
+    .. versionchanged:: 5.2
+       The *choices* parameter will now accept a sequence of tuples, where each
+       tuple is of form (*choice*, *description*)
+    """
+
+    def __init__(self, num_type, type_name,
+                 min=None, max=None, choices=None):
+        super(Number, self).__init__(type_name=type_name)
+
+        if min is not None and max is not None and max < min:
+            raise ValueError('Max value is less than min value')
+
+        if choices is not None:
+            if not all(isinstance(choice, tuple) for choice in choices):
+                choices = [(choice, None) for choice in choices]
+
+            self.choices = collections.OrderedDict(choices)
+        else:
+            self.choices = None
+
+        invalid_choices = [c for c in self.choices or []
+                           if (min is not None and min > c)
+                           or (max is not None and max < c)]
+        if invalid_choices:
+            raise ValueError("Choices %s are out of bounds [%s..%s]"
+                             % (invalid_choices, min, max))
+
+        self.min = min
+        self.max = max
+        self.num_type = num_type
+
+    def __call__(self, value):
+        if not isinstance(value, self.num_type):
+            s = str(value).strip()
+            if s == '':
+                return None
+            value = self.num_type(value)
+
+        if self.choices is None:
+            if self.min is not None and value < self.min:
+                raise ValueError('Should be greater than or equal to %g' %
+                                 self.min)
+            if self.max is not None and value > self.max:
+                raise ValueError('Should be less than or equal to %g' %
+                                 self.max)
+        else:
+            if value not in self.choices:
+                raise ValueError('Valid values are %r, but found %g' % (
+                                 self.choices, value))
+        return value
+
+    def __repr__(self):
+        props = []
+        if self.choices is not None:
+            props.append("choices={!r}".format(list(self.choices.keys())))
+        else:
+            if self.min is not None:
+                props.append('min=%g' % self.min)
+            if self.max is not None:
+                props.append('max=%g' % self.max)
+
+        if props:
+            return self.__class__.__name__ + '(%s)' % ', '.join(props)
+        return self.__class__.__name__
+
+    def __eq__(self, other):
+        return (
+            (self.__class__ == other.__class__) and
+            (self.min == other.min) and
+            (self.max == other.max) and
+            (set([x for x in self.choices or []]) ==
+                set([x for x in other.choices or []]) if
+             self.choices and other.choices else
+             self.choices == other.choices)
+        )
+
+    def _formatter(self, value):
+        return str(value)
+
+
+class Integer(Number):
 
     """Integer type.
 
@@ -249,12 +360,10 @@ class Integer(ConfigType):
     If value is whitespace or empty string will return None.
 
     :param min: Optional check that value is greater than or equal to min.
-                Mutually exclusive with 'choices'.
     :param max: Optional check that value is less than or equal to max.
-                Mutually exclusive with 'choices'.
     :param type_name: Type name to be used in the sample config file.
-    :param choices: Optional sequence of valid values. Mutually exclusive
-                    with 'min/max'.
+    :param choices: Optional sequence of either valid values or tuples of valid
+        values with descriptions.
 
     .. versionchanged:: 2.4
        The class now honors zero for *min* and *max* parameters.
@@ -264,108 +373,78 @@ class Integer(ConfigType):
 
     .. versionchanged:: 3.2
        Added *choices* parameter.
+
+    .. versionchanged:: 3.16
+       *choices* is no longer mutually exclusive with *min*/*max*. If those are
+       supplied, all choices are verified to be within the range.
+
+    .. versionchanged:: 5.2
+       The *choices* parameter will now accept a sequence of tuples, where each
+       tuple is of form (*choice*, *description*)
     """
 
     def __init__(self, min=None, max=None, type_name='integer value',
                  choices=None):
-        super(Integer, self).__init__(type_name=type_name)
-        if choices is not None:
-            if min is not None or max is not None:
-                raise ValueError("'choices' and 'min/max' cannot both be "
-                                 "specified")
-        else:
-            if min is not None and max is not None and max < min:
-                raise ValueError('Max value is less than min value')
-        self.min = min
-        self.max = max
-        self.choices = choices
-
-    def __call__(self, value):
-        if not isinstance(value, int):
-            s = str(value).strip()
-            if s == '':
-                value = None
-            else:
-                value = int(value)
-
-        if value is not None:
-            if self.choices is not None:
-                self._check_choices(value)
-            else:
-                self._check_range(value)
-
-        return value
-
-    def _check_choices(self, value):
-        if value in self.choices:
-            return
-        else:
-            raise ValueError('Valid values are %r, but found %d' % (
-                             self.choices, value))
-
-    def _check_range(self, value):
-        if self.min is not None and value < self.min:
-            raise ValueError('Should be greater than or equal to %d' %
-                             self.min)
-        if self.max is not None and value > self.max:
-            raise ValueError('Should be less than or equal to %d' % self.max)
-
-    def __repr__(self):
-        props = []
-        if self.choices is not None:
-            props.append("choices=%r" % (self.choices,))
-        else:
-            if self.min is not None:
-                props.append('min=%d' % self.min)
-            if self.max is not None:
-                props.append('max=%d' % self.max)
-
-        if props:
-            return 'Integer(%s)' % ', '.join(props)
-        return 'Integer'
-
-    def __eq__(self, other):
-        return (
-            (self.__class__ == other.__class__) and
-            (self.min == other.min) and
-            (self.max == other.max) and
-            (set(self.choices) == set(other.choices) if
-             self.choices and other.choices else
-             self.choices == other.choices)
-        )
-
-    def _formatter(self, value):
-        return str(value)
+        super(Integer, self).__init__(int, type_name, min=min, max=max,
+                                      choices=choices)
 
 
-class Float(ConfigType):
+class Float(Number):
 
     """Float type.
 
+    :param min: Optional check that value is greater than or equal to min.
+    :param max: Optional check that value is less than or equal to max.
     :param type_name: Type name to be used in the sample config file.
 
     .. versionchanged:: 2.7
 
        Added *type_name* parameter.
+
+    .. versionchanged:: 3.14
+
+       Added *min* and *max* parameters. If *choices* are also supplied, they
+       must be within the range.
     """
 
-    def __init__(self, type_name='floating point value'):
-        super(Float, self).__init__(type_name=type_name)
+    def __init__(self, min=None, max=None, type_name='floating point value'):
+        super(Float, self).__init__(float, type_name, min=min, max=max)
 
-    def __call__(self, value):
-        if isinstance(value, float):
-            return value
 
-        return float(value)
+class Port(Integer):
 
-    def __repr__(self):
-        return 'Float'
+    """Port type
 
-    def __eq__(self, other):
-        return self.__class__ == other.__class__
+    Represents a L4 Port.
 
-    def _formatter(self, value):
-        return str(value)
+    :param min: Optional check that value is greater than or equal to min.
+    :param max: Optional check that value is less than or equal to max.
+    :param type_name: Type name to be used in the sample config file.
+    :param choices: Optional sequence of either valid values or tuples of valid
+        values with descriptions.
+
+    .. versionadded:: 3.16
+
+    .. versionchanged:: 5.2
+       The *choices* parameter will now accept a sequence of tuples, where each
+       tuple is of form (*choice*, *description*)
+    """
+
+    PORT_MIN = 0
+    PORT_MAX = 65535
+
+    def __init__(self, min=None, max=None, type_name='port', choices=None):
+        min = self.PORT_MIN if min is None else min
+        max = self.PORT_MAX if max is None else max
+        if min < self.PORT_MIN:
+            raise ValueError('Min value cannot be less than %(min)d' %
+                             {'min': self.PORT_MIN})
+        if max > self.PORT_MAX:
+            raise ValueError('Max value cannot be more than %(max)d' %
+                             {'max': self.PORT_MAX})
+
+        super(Port, self).__init__(min=min, max=max, type_name=type_name,
+                                   choices=choices)
 
 
 class List(ConfigType):
@@ -380,7 +459,8 @@ class List(ConfigType):
     comma and next item until validation succeeds or there is no parts left.
     In the later case it will signal validation error.
 
-    :param item_type: type of list items
+    :param item_type: Type of list items. Should be an instance of
+                      ``ConfigType``.
     :param bounds: if True, value should be inside "[" and "]" pair
     :param type_name: Type name to be used in the sample config file.
 
@@ -401,23 +481,24 @@ class List(ConfigType):
         self.bounds = bounds
 
     def __call__(self, value):
-        if isinstance(value, list):
-            return value
-
-        result = []
-        s = value.strip()
+        if isinstance(value, (list, tuple)):
+            return list(map(self.item_type, value))
 
+        s = value.strip().rstrip(',')
         if self.bounds:
             if not s.startswith('['):
                 raise ValueError('Value should start with "["')
             if not s.endswith(']'):
                 raise ValueError('Value should end with "]"')
             s = s[1:-1]
+        if s:
+            values = s.split(',')
+        else:
+            values = []
+        if not values:
+            return []
 
-        if s == '':
-            return result
-
-        values = s.split(',')
+        result = []
         while values:
             value = values.pop(0)
             while True:
@@ -447,7 +528,72 @@ class List(ConfigType):
         )
 
     def _formatter(self, value):
-        return ','.join(value)
+        fmtstr = '[{}]' if self.bounds else '{}'
+        if isinstance(value, str):
+            return fmtstr.format(value)
+        if isinstance(value, list):
+            value = [
+                self.item_type._formatter(v)
+                for v in value
+            ]
+            return fmtstr.format(','.join(value))
+        return fmtstr.format(self.item_type._formatter(value))
+
+
+class Range(ConfigType):
+
+    """Range type.
+
+    Represents a range of integers. A range is identified by an integer both
+    sides of a '-' character. Negatives are allowed. A single number is also a
+    valid range.
+
+    :param min: Optional check that lower bound is greater than or equal to
+                min.
+    :param max: Optional check that upper bound is less than or equal to max.
+    :param inclusive: True if the right bound is to be included in the range.
+    :param type_name: Type name to be used in the sample config file.
+
+    .. versionadded:: 3.18
+    """
+
+    def __init__(self, min=None, max=None, inclusive=True,
+                 type_name='range value'):
+        super(Range, self).__init__(type_name)
+        self.min = min
+        self.max = max
+        self.inclusive = inclusive
+
+    def __call__(self, value):
+        value = str(value)
+        num = "0|-?[1-9][0-9]*"
+        m = re.match("^(%s)(?:-(%s))?$" % (num, num), value)
+        if not m:
+            raise ValueError('Invalid Range: %s' % value)
+        left = int(m.group(1))
+        right = int(left if m.group(2) is None else m.group(2))
+
+        if left < right:
+            left = Integer(min=self.min)(left)
+            right = Integer(max=self.max)(right)
+            step = 1
+        else:
+            left = Integer(max=self.max)(left)
+            right = Integer(min=self.min)(right)
+            step = -1
+        if self.inclusive:
+            right += step
+        return range(left, right, step)
+
+    def __eq__(self, other):
+        return (
+            (self.__class__ == other.__class__) and
+            (self.min == other.min) and
+            (self.max == other.max)
+        )
+
+    def _formatter(self, value):
+        return value
 
 
 class Dict(ConfigType):
@@ -606,7 +752,7 @@ class IPAddress(ConfigType):
 
 
 class Hostname(ConfigType):
-    """Hostname type.
+    """Host domain name type.
 
     A hostname refers to a valid DNS or hostname. It must not be longer than
     253 characters, have a segment greater than 63 characters, nor start or
@@ -615,21 +761,26 @@ class Hostname(ConfigType):
     :param type_name: Type name to be used in the sample config file.
 
     """
+    HOSTNAME_REGEX = '(?!-)[A-Z0-9-]{1,63}(?<!-)$'
 
     def __init__(self, type_name='hostname value'):
         super(Hostname, self).__init__(type_name=type_name)
 
-    def __call__(self, value):
+    def __call__(self, value, regex=HOSTNAME_REGEX):
         """Check hostname is valid.
 
         Ensures that each segment
         - Contains at least one character and a maximum of 63 characters
         - Consists only of allowed characters: letters (A-Z and a-z),
           digits (0-9), and hyphen (-)
+        - Ensures that the final segment (representing the top level domain
+          name) contains at least one non-numeric character
         - Does not begin or end with a hyphen
         - maximum total length of 253 characters
 
-        For more details , please see: http://tools.ietf.org/html/rfc1035
+        For more details , please see: http://tools.ietf.org/html/rfc1035,
+        https://www.ietf.org/rfc/rfc1912, and
+        https://tools.ietf.org/html/rfc1123
         """
 
         if len(value) == 0:
@@ -639,7 +790,11 @@ class Hostname(ConfigType):
                              % value)
         if value.endswith("."):
             value = value[:-1]
-        allowed = re.compile("(?!-)[A-Z0-9-]{1,63}(?<!-)$", re.IGNORECASE)
+        allowed = re.compile(regex, re.IGNORECASE)
+        if not re.search('[a-zA-Z-]', value.split(".")[-1]):
+            raise ValueError('%s contains no non-numeric characters in the '
+                             'top-level domain part of the host name and is '
+                             'invalid' % value)
         if any((not allowed.match(x)) for x in value.split(".")):
             raise ValueError("%s is an invalid hostname" % value)
         return value
@@ -652,3 +807,170 @@ class Hostname(ConfigType):
 
     def _formatter(self, value):
         return value
+
+
+class HostAddress(ConfigType):
+    """Host Address type.
+
+    Represents both valid IP addresses and valid host domain names
+    including fully qualified domain names.
+    Performs strict checks for both IP addresses and valid hostnames,
+    matching the opt values to the respective types as per RFC1912.
+
+    :param version: defines which version should be explicitly
+                    checked (4 or 6) in case of an IP address
+    :param type_name: Type name to be used in the sample config file.
+    """
+
+    def __init__(self, version=None, type_name='host address value'):
+        """Check for valid version in case an IP address is provided
+
+        """
+
+        super(HostAddress, self).__init__(type_name=type_name)
+        self.ip_address = IPAddress(version, type_name)
+        self.hostname = Hostname('localhost')
+
+    def __call__(self, value):
+        """Checks if is a valid IP/hostname.
+
+        If not a valid IP, makes sure it is not a mistyped IP before
+        performing checks for it as a hostname.
+
+        """
+
+        try:
+            value = self.ip_address(value)
+        except ValueError:
+            try:
+                value = self.hostname(value)
+            except ValueError:
+                raise ValueError(
+                    "%s is not a valid host address" % (value,))
+        return value
+
+    def __repr__(self):
+        return 'HostAddress'
+
+    def __eq__(self, other):
+        return self.__class__ == other.__class__
+
+    def _formatter(self, value):
+        return value
+
+
+class HostDomain(HostAddress):
+    """Host Domain type.
+
+    Like HostAddress with the support of _ character.
+
+    :param version: defines which version should be explicitly
+                    checked (4 or 6) in case of an IP address
+    :param type_name: Type name to be used in the sample config file.
+    """
+    # DOMAIN_REGEX is HOSTNAME_REGEX with the _ character added
+    DOMAIN_REGEX = '(?!-)[A-Z0-9-_]{1,63}(?<!-)$'
+
+    def __call__(self, value):
+        """Checks if is a valid IP/hostname.
+
+        If not a valid IP, makes sure it is not a mistyped IP before
+        performing checks for it as a hostname.
+
+        """
+
+        try:
+            value = super(HostDomain, self).__call__(value)
+        except ValueError:
+            # Check if domain is valid
+            # Add support of underscore
+            # https://www.ietf.org/rfc/rfc1912,
+            # http://domainkeys.sourceforge.net/underscore.html
+            # https://bugs.launchpad.net/oslo.config/+bug/1892044
+            try:
+                value = self.hostname(value, regex=self.DOMAIN_REGEX)
+            except ValueError:
+                raise ValueError(
+                    "%s is not a valid host address" % (value,))
+        return value
+
+    def __repr__(self):
+        return 'HostDomain'
+
+
+class URI(ConfigType):
+
+    """URI type
+
+    Represents URI. Value will be validated as RFC 3986.
+
+    :param max_length: Optional integer. If a positive value is specified,
+                       a maximum length of an option value must be less than
+                       or equal to this parameter. Otherwise no length check
+                       will be done.
+    :param schemes: List of valid schemes.
+    :param type_name: Type name to be used in the sample config file.
+
+    .. versionchanged:: 3.14
+       Added *max_length* parameter.
+    .. versionchanged:: 3.18
+       Added *schemes* parameter.
+    """
+
+    def __init__(self, max_length=None, schemes=None, type_name='uri value'):
+        super(URI, self).__init__(type_name=type_name)
+        self.max_length = max_length
+        self.schemes = schemes
+
+    def __call__(self, value):
+        uri = rfc3986.uri_reference(value)
+        validator = rfc3986.validators.Validator().require_presence_of(
+            'scheme', 'host',
+        ).check_validity_of(
+            'scheme', 'host', 'path',
+        )
+        if self.schemes:
+            validator = validator.allow_schemes(*self.schemes)
+        try:
+            validator.validate(uri)
+        except rfc3986.exceptions.RFC3986Exception as exc:
+            raise ValueError(exc)
+
+        if self.max_length is not None and len(value) > self.max_length:
+            raise ValueError("Value '%s' exceeds maximum length %d" %
+                             (value, self.max_length))
+
+        # NOTE(dhellmann): self.value is deprecated, and we don't want
+        # to trigger a deprecation warning ourselves so we modify
+        # self._value directly.
+        self._value = value
+        return value
+
+    @removals.removed_property
+    def value(self):
+        return self._value
+
+    @value.setter
+    def value(self, newval):
+        self._value = newval
+
+    @value.deleter
+    def value(self):
+        del self._value
+
+    def __repr__(self):
+        return 'URI'
+
+    def __eq__(self, other):
+        to_compare = ['__class__', 'max_length', 'schemes']
+        unset = object()
+        my_values = tuple(
+            getattr(self, name, unset) for name in to_compare
+        )
+        other_values = tuple(
+            getattr(other, name, unset) for name in to_compare
+        )
+        return my_values == other_values
+
+    def _formatter(self, value):
+        return value
